#include "RenderTraverser.h"
#include "glInterface.h"
#include "Timers.h"
#include "Camera.h"
#include "SceneEntity.h"
#include "RenderState.h"


namespace CHCDemo
{

	
void TraversalStatistics::Reset()
{
	// reset statistics
	mNumTraversedNodes = 0;
	mNumQueryCulledNodes = 0;
	mNumFrustumCulledNodes = 0;
	mNumRenderedGeometry = 0;
	mNumRenderedTriangles = 0;

	mNumIssuedQueries = 0;
	mRenderTime = 0;
}



/******************************************************/
/*           RenderTraverser implementation           */
/******************************************************/


RenderTraverser::RenderTraverser():  
mVisibilityThreshold(0),
mBvh(NULL), 
mIsQueryMode(false),
mUseOptimization(true),
mFrameId(-1),
mUseRenderQueue(false)
{
}


RenderTraverser::~RenderTraverser()
{
	mQueryHandler.DestroyQueries();
}


void RenderTraverser::EnqueueNode(BvhNode *node)
{
	mBvh->CalcDistance(node);
	mDistanceQueue.push(node);
}


void RenderTraverser::TraverseNode(BvhNode *node)
{
	++ mStats.mNumTraversedNodes;

	if (node->IsVirtualLeaf())
	{
		if (mUseRenderQueue)
		{
			int geometrySize;
			SceneEntity **geom = mBvh->GetGeometry(node, geometrySize);
			mRenderQueue.Enqueue(geom, geometrySize);
			mStats.mNumRenderedGeometry += geometrySize;
		}
		else
		{
			// RenderBox(node->GetBox());
			mStats.mNumRenderedGeometry += mBvh->Render(node, mRenderState);
		}		
	}
	else 
	{
		// for non leafs this renders only the bounding volume (if the flag is set)
		BvhInterior *interior = static_cast<BvhInterior *>(node);

		EnqueueNode(interior->GetFront());
		EnqueueNode(interior->GetBack());
	}
}


void RenderTraverser::SetHierarchy(Bvh *bvh)
{
	mBvh = bvh;
}


void RenderTraverser::SetRenderState(RenderState *state)
{
	mRenderState = state;
	mRenderQueue.SetRenderState(state);
}


OcclusionQuery *RenderTraverser::IssueOcclusionQuery(BvhNode *node, bool wasVisible)
{
	++ mStats.mNumIssuedQueries;

	OcclusionQuery *query = mQueryHandler.RequestQuery();
	query->Reset();
	query->AddNode(node);

	query->BeginQuery();
	
	// if leaf and was visible => will be rendered anyway, thus we
	// can also test with the real geometry 
	/*if(node->IsLeaf() && wasVisible && mUseOptimization)
	{
		mNumRenderedGeometry += node->Render();
	}
	else
	{*/

	// change to query mode
	mRenderState->SetState(RenderState::QUERY);
	mBvh->RenderBoundingBox(node);
	mRenderState->SetState(RenderState::RENDER);

	query->EndQuery();

	return query;
}


void RenderTraverser::SetVisibilityThreshold(int threshold)
{
	mVisibilityThreshold = threshold;
}


int RenderTraverser::GetVisibilityThreshold() const
{
	return mVisibilityThreshold;
}


void RenderTraverser::SetUseOptimization(bool useOptimization)
{
	mUseOptimization = useOptimization;
}



void RenderTraverser::RenderScene()
{
	InitTiming();
	long t1, t2;

	t1 = GetTime();

	glEnableClientState(GL_VERTEX_ARRAY);
	glEnableClientState(GL_NORMAL_ARRAY);

	++ mFrameId;

	mBvh->InitFrame(mCamera, mFrameId);

	mStats.Reset();
	mQueryHandler.ResetQueries();
	
	EnqueueNode(mBvh->GetRoot());


	///////////
	//-- the actual rendering algorithm

	Traverse();

	glDisableClientState(GL_VERTEX_ARRAY);
	glDisableClientState(GL_NORMAL_ARRAY);

	mRenderState->Reset();

	t2 = GetTime();
	mStats.mRenderTime = TimeDiff(t1, t2);
}


void RenderTraverser::SetUseRenderQueue(bool useRenderQueue)
{
	mUseRenderQueue = useRenderQueue;
	std::cout << "using render queue: " << mUseRenderQueue << std::endl;
}

}
